# coding: utf-8
#
# Copyright (c) 2024 National Technology and Engineering Solutions of
# Sandia, LLC. Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# SPDX-License-Identifier: BSD-3-Clause.

"""
The SANSMIC main program.

This submodule provides the hook for the command line programs ``sansmic`` 
and ``sansmic-convert``.

"""

import logging
from argparse import ArgumentParser
from os.path import splitext

from sansmic import __version__, io

logging.basicConfig()
logger = logging.getLogger("sansmic")


def main():
    """Command line function to run sansmic."""
    parser = ArgumentParser(
        prog="sansmic",
        description=f"A solution mining code. v{__version__}",
    )
    parser.add_argument("datafile", help="the SANSMIC input file to use")
    parser.add_argument(
        "-o",
        "--output",
        dest="prefix",
        help="change result files' prefix [default: from datafile]",
        default=None,
    )
    verb = parser.add_mutually_exclusive_group()
    verb.add_argument(
        "-v",
        "--verbose",
        help="Increase the output verbosity.",
        action="count",
        default=0,
    )
    verb.add_argument(
        "-q",
        "--quiet",
        help="Only report errors, not warnings.",
        action="store_true",
        default=False,
    )
    args = parser.parse_args()
    datafile = args.datafile
    prefix = splitext(datafile)[0] if args.prefix is None else args.prefix
    level = logging.WARNING
    if args.verbose >= 2:
        level = logging.DEBUG
    elif args.verbose:
        level = logging.INFO
    elif args.quiet:
        level = logging.ERROR
    logger.setLevel(level)
    logger.debug("Running sansmic with {}".format(args))
    model = io.read_scenario(datafile, warn=not args.quiet)
    logger.debug("Successfully created scenario from {}".format(datafile))
    logger.info("Running simulation")
    with model.new_simulation(prefix) as sim:
        sim.run()


def convert():
    """Command line function to convert a scenario/dat to a new format."""
    parser = ArgumentParser(
        prog="sansmic-convert",
        description=f"A solution mining code. v{__version__}",
    )
    parser.add_argument("infile", help="the original input file to convert")
    parser.add_argument(
        dest="outfile",
        help="the new scenario file to create",
        default=None,
    )
    verb = parser.add_mutually_exclusive_group()
    verb.add_argument(
        "-v",
        "--verbose",
        help="increase the output verbosity: once -> info, twice -> debug.",
        action="count",
        default=0,
    )
    verb.add_argument(
        "-q",
        "--quiet",
        help="only report errors, not warnings.",
        action="store_true",
        default=False,
    )
    args = parser.parse_args()
    infile = args.infile
    level = logging.WARNING
    if args.verbose >= 2:
        level = logging.DEBUG
    elif args.verbose:
        level = logging.INFO
    elif args.quiet:
        level = logging.ERROR
    logger.setLevel(level)
    logger.debug("Running sansmic-convert")
    model = io.read_scenario(infile, warn=not args.quiet)
    logger.debug("Successfully created scenario from {}".format(infile))
    io.write_scenario(model, args.outfile)
    logger.debug("Successfully wrote scenario to {}".format(args.outfile))



# coding: utf-8
#
# Copyright (c) 2024 National Technology and Engineering Solutions of
# Sandia, LLC. Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# SPDX-License-Identifier: BSD-3-Clause.

"""The core python classes for SANSMIC."""

import warnings
from dataclasses import InitVar, asdict, dataclass, field
from enum import IntEnum
from fractions import Fraction
import logging
from math import nan
from types import NoneType
from typing import Any, Dict, List, Union

import numpy as np
import pandas as pd

logger = logging.getLogger("sansmic")

has_ext = False
try:
    from . import libsansmic as _ext
    has_ext = True
except ImportError:
    logger.critical('The C++ library is not installed. Conversions will work, but the main program will not run.')
    has_ext = False
    class _ext:
        class _CModel: pass
        class _CStage: pass
        class _CResults: pass
        class _CGeometryFormat(IntEnum):
            blah = 1
        class _CRunMode(IntEnum):
            blah = 1
        CStage = _CStage
        CModel = _CModel
        CGeometryFormat = _CGeometryFormat
        CRunMode = _CRunMode
        CResults = _CResults


class SansmicConfigError(TypeError):
    """Specific error for missing configuration data."""

    pass


class Units(IntEnum):
    """The units that are used to define the scenario.

    Depths, heights, and cavern radii are in the first (larger) length unit.
    Tubing radii are in the second (smaller) length unit.
    Volumes and volumetric flow rates are in the third unit.

    Durations are in hours. Constant injection rates are in <volume-unit> per day.
    File-based injection rates are in <volume-unit> per hour.
    """

    FT_IN_BBL = 1
    """foot/inch/barrel"""
    FT_IN_FT3 = 2
    """foot/inch/cubic foot"""
    M_CM_M3 = 3
    """meter/centimeter/cubic meter"""

    @property
    def inch(self):
        """1 in ≝ 0.0254 m"""
        return Fraction(254, 10000)

    @property
    def foot(self):
        """1 ft ≝ 0.3048 m"""
        return Fraction(3048, 10000)

    @property
    def cubic_foot(self):
        """1 ft³ == 0.028316846592 m³"""
        return Fraction(3048**3, 10000**3)

    @property
    def barrel(self):
        """1 bbl == 0.158987294928 m³"""
        return Fraction(42 * 231 * 254, 10000**3)

    @property
    def centimeter(self):
        """1 cm ≝ 0.01 m"""
        return Fraction(1, 100)


class StopCondition(IntEnum):
    """Which stop condition should be used to end the simulation."""

    DEPTH = -1
    """Stop when the interface reaches a specified height above the cavern floor"""
    DURATION = 0
    """Stop only based on duration"""
    VOLUME = 1
    """Stop when the total cavern volume reaches a specified value."""


class GeometryFormat(IntEnum):
    """The format for the initial cavern geometry."""

    RADIUS_LIST = 0
    """The radius is provided from the bottom of the cavern to the top;
    :attr:`~Scenario.num_cells` + 1 values, equally spaced"""
    VOLUME_LIST = 1
    VOLUME_TABLE = -1
    RADIUS_TABLE = 2
    LAYER_CAKE = 5
    """The geometry is provided in a 'layer-cake' style LAS file"""

class SimulationMode(IntEnum):
    """The simulation mode determines which options are active for injection."""

    ORDINARY = 0
    """Ordinary leaching, with raw water or undersaturated brine injected through
    the inner tubing or outer casing and brine produced from the other."""
    WITHDRAWAL = 1
    """Withdrawal leach, with brine injected through the suspended tubing (hanging 
    string) and product produced from top of cavern."""
    LEACH_FILL = 2
    """Simultaneous leaching (water/brine injection and brine production) and 
    product injection from the top."""
    STORAGE_FILL = -1
    """Storage fill, with product injected from the top of the cavern and brine
    produced from the suspended tubing/hanging string."""


class RateScheduleType(IntEnum):
    """The way that the rate of injection is specified."""

    CONSTANT_RATE = 1
    """Injection occurs at a constant rate of volume :class:`Units` per day."""
    DATAFILE = 2
    """Injection rate is specified in :class:`Units` per hour in a file."""


@dataclass
class StageDefinition:
    """Defines a SANSMIC simulation stage which is an injection period followed by a rest period."""

    title: str = None
    """Title for the stage."""
    simulation_mode: SimulationMode = None
    """The simulation mode used in this stage."""
    solver_timestep: float = None
    """The solver timestep in hours."""
    save_frequency: int = None
    """The save frequency in number of timesteps"""
    injection_duration: float = None
    """The duration of the injection phase of the stage."""
    rest_duration: float = None
    """The duration of the post-injection rest phase of the stage."""

    inner_tbg_inside_diam: float = None
    """The inner tubing inner diameter (ID)."""
    inner_tbg_outside_diam: float = None
    """The inner tubing outer diameter (OD)."""
    outer_csg_inside_diam: float = None
    """The outer tubing/casing inner diameter (ID)."""
    outer_csg_outside_diam: float = None
    """The outer tubing/casing outer diameter (OD)."""

    brine_injection_sg: float = None
    """The water/brine injection specific gravity."""
    brine_injection_depth: float = None
    """The depth below surface/:term:`ZDP` where the end of the injection string/casing/tubing is positioned."""
    brine_production_depth: float = None
    """The depth below surface/:term:`ZDP` where the end of the production string/casing/tubing is positioned."""
    brine_injection_rate: Union[float, str] = 0
    """The meaning of this field is based on the type of data that is provided (a value of 0 means off).

    float : constant rate
        Injection occurs at a constant rate for the duration of the
        :attr:`injection_duration`.
    
    str : file path
        Injection rates should be read from a file. The file can be any type that can
        be read into a pd DataFrame. An 'index' column is optional, but if not provided, then all data is assumed to be pre-sorted. In addition to the 
        'index' column, files may contain exactly one set of the following columns:
        
        ['rate', 'duration']
            This style gives a list of rates and how long they should continue for.
            Durations are provided in decimal hours. If the sum of durations is greater than :attr:`injection_duration`, an error will be raised. If the 
            sum of durations is less than :attr:`injection_duration`, the remaining
            time will be assumed to be no-flow.
        
        ['time', 'rate']
            This style gives a list of new rate values that start at the given time 
            after the start of the stage. The 'time' values must be increasing and
            never duplicate. Times are in decimal hours since the start of the stage.
            The last rate will be used up until the end of the 
            :attr:`injection_duration`. If a time is given that is after the 
            :attr:`injection_duration`, then an error will be raised.

        ['hourly']
            This style specifies a new rate for each hour of the stage. There must be
            a value specified for each hour of the :attr`injection_duration`, or an 
            error will be raised.

        The optional column ['sg'] can be provided to change the injection water/brine
        specific gravity; however, if provided, it must contain a value for every row.

        
        If 'duration' and 'time' occur in the same file, an error will be raised.
        If 'time', 'rate', or 'duration' occur in the same file as 'hourly', then
        an error will be raised. 
        
        Columns with names other than those described above will be ignored.

        .. attention::

            In all cases, a no-flow, leaching-only period will occur for
            :attr:`rest_duration` hours **after** all injection data
            is processed -- this means that if an injection file finishes with 0s
            that there will be an additional :attr:`rest_duration` hours added to the 
            end of the stage.

        .. warning::

            The same file cnnot be used for both brine injection and product
            injection data.
    """

    set_initial_conditions: bool = None
    """Unlink initial cavern brine gravity and interface level from previous stage. 
    Automatically set to True for the first stage added to a model."""
    set_cavern_sg: float = None
    """Set the initial specific gravity for all brine-filled cells of the cavern to 
    this value. If set_initial_conditions is False and this is not None (or 0) an error
    will be raised upon scenario validation."""
    set_interface_level: float = None
    """Set the initial oil-brine interface or blanket level. By default None, which
    will link to the previous stage (as will a value of 0)."""
    # set_cavern_temp: float = None
    # """Set the initial temperature for all brine-filled cells of the cavern; by
    # default None, which will link to the previous stage."""

    product_injection_rate: Union[float, str] = 0.0
    """Either a constant rate of product injection or a file with an injection schedule, by default 0."""
    product_water_content: float = None
    """The volume-percent water within the product; a value of None turns this option off."""
    product_injection_depth: float = None
    """The depth below surface/:term:`ZDP` where the product injection string -- or casing or chimney bottom -- is positioned; a value 0 or None sets this to the roof of the cavern."""
    product_production_depth: float = None
    """The depth below surface/:term:`ZDP` where the product production string -- or casing or chimney bottom -- is positioned; a value 0 or None sets this to the roof of the cavern."""

    stop_condition: StopCondition = StopCondition.DURATION
    """Set an additional stop condition for the stage, see :class:`StopCondition`."""
    stop_value: float = 0
    """If the :attr:`stop_condition` is not :attr:`~StopCondition.DURATION`, then the depth or volume value."""

    defaults: InitVar[Dict[str, Any]] = None

    valid_default_keys = [
        "solver_timestep",
        "save_frequency",
        "inner_tbg_inside_diam",
        "inner_tbg_outside_diam",
        "outer_csg_inside_diam",
        "outer_csg_outside_diam",
    ]

    def __post_init__(self, defaults=None):
        if defaults is None:
            return
        if not isinstance(defaults, dict):
            raise TypeError("defaults must be a dictionary")
        for k, v in defaults.items():
            if k not in self.valid_default_keys:
                logger.warning(
                    "Ignoring non-defaultable or unknown setting {} = {}".format(k, repr(v))
                )
            elif getattr(self, k) is None:
                setattr(self, k, v)

    def __setattr__(self, name, value):
        if isinstance(value, str) and value.strip() == "":
            value = None
        if name == "simulation_mode" and not isinstance(value, SimulationMode):
            if isinstance(value, int):
                value = SimulationMode(value)
            elif isinstance(value, str):
                value = SimulationMode[value.upper().replace(" ", "_").replace("-", "_")]
            elif isinstance(value, _ext.CRunMode):
                value = SimulationMode(int(value))
            else:
                TypeError("simulation_mode cannot be of type {}".format(type(value)))
        elif name == "stop_condition" and not isinstance(value, StopCondition):
            if isinstance(value, int):
                value = StopCondition(value)
            elif isinstance(value, str):
                value = StopCondition[value.upper().replace(" ", "_").replace("-", "_")]
            elif isinstance(value, _ext.CGeometryFormat):
                value = StopCondition(int(value))
            else:
                SansmicConfigError("stop_condition cannot be of type {}".format(type(value)))
        elif name == "set_cavern_sg" and value is not None and value < 1.0:
            value = None
        elif name == "set_initial_conditions" and value is False:
            self.set_cavern_sg = None
            self.set_interface_level = None
        super().__setattr__(name, value)

    def validate(self):
        """Validate that all required options have been selected.

        Raises
        ------
        SansmicConfigError
            if there are missing or misconfigured options
        ValueError
            if timing options are invalid
        """
        if self.solver_timestep <= 0:
            raise ValueError("Timestep must be greater than 0 hours")
        if self.injection_duration <= 0:
            raise ValueError("Injection duration must be greater than 0 hours")
        if self.rest_duration < 0:
            raise ValueError("Rest duration must be positive")

        # Validate appropriate initial conditions settings
        if self.set_initial_conditions and (
            self.set_cavern_sg is None or self.set_interface_level is None
        ):
            raise SansmicConfigError(
                "An initial stage must have both set_cavern_sg and set_interface_level."
            )
        elif not self.set_initial_conditions and self.set_cavern_sg:
            # raise SansmicConfigError(
            warnings.warn(
                "Setting the starting cavern sg ought to use 'set_initial_conditions' to be set to True"
            )
        elif not self.set_initial_conditions and self.set_interface_level:
            warnings.warn(
                "Make sure you meant to reset the interface level; use 0.0 or None to continue from the last stage."
            )

        # Validate appropriate brine injection and production settings
        if (
            self.simulation_mode
            in [SimulationMode.ORDINARY, SimulationMode.LEACH_FILL, SimulationMode.STORAGE_FILL]
            and self.brine_production_depth is None
        ):
            raise SansmicConfigError(
                "Missing required 'brine_production_depth' for {} simulation mode.".format(
                    self.simulation_mode.name
                )
            )
        if (
            self.simulation_mode
            in [SimulationMode.ORDINARY, SimulationMode.LEACH_FILL, SimulationMode.WITHDRAWAL]
            and self.brine_injection_depth is None
        ):
            raise SansmicConfigError(
                "Missing required 'brine_injection_depth' for {} simulation mode.".format(
                    self.simulation_mode.name
                )
            )
        if (
            self.simulation_mode
            in [SimulationMode.ORDINARY, SimulationMode.LEACH_FILL, SimulationMode.WITHDRAWAL]
            and self.brine_injection_rate is None
        ):
            raise SansmicConfigError(
                "Missing required 'brine_injection_rate' for {} simulation mode.".format(
                    self.simulation_mode.name
                )
            )
        if (
            self.simulation_mode
            in [SimulationMode.ORDINARY, SimulationMode.LEACH_FILL, SimulationMode.WITHDRAWAL]
            and self.brine_injection_sg is None
        ):
            raise SansmicConfigError(
                "Missing required 'brine_injection_sg' for {} simulation mode.".format(
                    self.simulation_mode.name
                )
            )

        # Validate casing size information
        if (
            self.simulation_mode
            in [
                SimulationMode.ORDINARY,
                SimulationMode.LEACH_FILL,
                SimulationMode.WITHDRAWAL,
                SimulationMode.STORAGE_FILL,
            ]
            and self.inner_tbg_inside_diam is None
        ):
            raise SansmicConfigError(
                "Missing required 'inner_tbg_inside_diam' for {} simulation mode.".format(
                    self.simulation_mode.name
                )
            )
        if (
            self.simulation_mode
            in [
                SimulationMode.ORDINARY,
                SimulationMode.LEACH_FILL,
                SimulationMode.WITHDRAWAL,
                SimulationMode.STORAGE_FILL,
            ]
            and self.inner_tbg_outside_diam is None
        ):
            raise SansmicConfigError(
                "Missing required 'inner_tbg_outside_diam' for {} simulation mode.".format(
                    self.simulation_mode.name
                )
            )
        if (
            self.simulation_mode in [SimulationMode.ORDINARY, SimulationMode.LEACH_FILL]
            and self.outer_csg_inside_diam is None
        ):
            raise SansmicConfigError(
                "Missing required 'outer_csg_inside_diam' for {} simulation mode.".format(
                    self.simulation_mode.name
                )
            )
        if (
            self.simulation_mode in [SimulationMode.ORDINARY, SimulationMode.LEACH_FILL]
            and self.outer_csg_outside_diam is None
        ):
            raise SansmicConfigError(
                "Missing required 'outer_csg_outside_diam' for {} simulation mode.".format(
                    self.simulation_mode.name
                )
            )

        # Product injection depth an rate validation
        if (
            self.simulation_mode in [SimulationMode.LEACH_FILL, SimulationMode.STORAGE_FILL]
            and self.product_injection_depth is None
        ):
            self.product_injection_depth = 0.0
            warnings.warn(
                "Missing 'product_injection_depth' options for {} simulation mode. Setting to 0.0".format(
                    self.simulation_mode.name
                )
            )
        if (
            self.simulation_mode in [SimulationMode.LEACH_FILL, SimulationMode.STORAGE_FILL]
            and self.product_injection_rate is None
        ):
            raise SansmicConfigError(
                "Missing required 'product_injection_rate' options for {} simulation mode.".format(
                    self.simulation_mode.name
                )
            )

    def to_dict(self) -> Dict[str, Any]:
        """Convert the stage definition to a dictionary"""
        ret = asdict(self)
        return ret

    def _to_cstage(self) -> _ext.CStage:
        """Create a CStage object for the C++ interface."""
        self.validate()
        stage = _ext.CStage()
        stage.title = self.title
        stage.mode = _ext.CRunMode(int(self.simulation_mode))
        stage.print_interval = self.save_frequency
        stage.subsequent = 0 if self.set_initial_conditions else 1
        stage.iResetGeo = 0
        stage.rest_duration = self.rest_duration
        stage.stop_value = (
            int(self.stop_condition) * abs(self.stop_value) if self.stop_condition else 0
        )
        stage.injection_depth = (
            self.brine_injection_depth if self.brine_injection_depth is not None else 0
        )
        stage.production_depth = (
            self.brine_production_depth if self.brine_production_depth is not None else 0
        )
        stage.interface_depth = (
            self.set_interface_level if self.set_interface_level is not None else 0
        )
        stage.injection_rate = self.brine_injection_rate
        stage.inn_tbg_inside_radius = self.inner_tbg_inside_diam / 2.0
        stage.inn_tbg_outside_radius = self.inner_tbg_outside_diam / 2.0
        stage.out_csg_inside_radius = (
            self.outer_csg_inside_diam / 2.0
            if self.outer_csg_inside_diam
            else self.inner_tbg_inside_diam / 2.0
        )
        stage.out_csg_outside_radius = (
            self.outer_csg_outside_diam / 2.0
            if self.outer_csg_outside_diam
            else self.inner_tbg_outside_diam / 2.0
        )
        stage.injection_fluid_sg = self.brine_injection_sg
        stage.cavern_sg = (
            0.0
            if not self.set_initial_conditions or self.set_cavern_sg is None
            else self.set_cavern_sg
        )
        stage.timestep = self.solver_timestep
        stage.injection_duration = self.injection_duration
        stage.fill_rate = (
            self.product_injection_rate if self.product_injection_rate is not None else 0.0
        )
        stage.tDelay = 0.0
        return stage


@dataclass
class Scenario:
    """A SANSMIC scenario definition used to run a simulation."""

    title: str = None
    """General title for the scenario."""
    num_cells: int = None
    """Number of cells to use in the model."""
    geometry_data: Union[List[float], List[List[float]], str] = None
    """The geometry data in a list or list of lists, or a filename."""
    geometry_format: GeometryFormat = GeometryFormat.RADIUS_LIST
    """The format for the geometry data."""
    cavern_height: float = None
    """The total cavern height."""
    floor_depth: float = None
    """The depth from surface datum of the floor."""
    ullage_standoff: float = 40
    """The standoff distance of the lowest brine interface depth."""
    insolubles_ratio: float = 0.04
    """The volume ratio of insoluble material within the salt."""
    dissolution_factor: float = 1.0
    """**Danger** modify the dissolution factor by a certain amount; leave this as 1.0."""
    coallescing_wells: int = 1
    """Number of coallescing wells for cavern development modeling; leave as 1 for completed caverns."""
    well_separation: float = 0.0
    """The separation distance between wells for cavern development modeling."""
    units: Units = Units.FT_IN_BBL
    """The units used in describing the scenario."""
    stages: List[StageDefinition] = field(default_factory=list)
    """The activity stages to simulate."""

    defaults: Dict[str, Union[int, float]] = field(default_factory=dict)
    """Default values for a subset of stage attributes, see :attr:`StageDefinition.valid_default_keys`."""

    def __post_init__(self):
        converted_stages = list()
        for stage in self.stages:
            if isinstance(stage, dict):
                stage = StageDefinition(**stage, defaults=self.defaults)
            converted_stages.append(stage)
        self.stages = converted_stages

    def __setattr__(self, name, value):
        if isinstance(value, str) and value.strip() == "":
            value = None
        if name == "geometry_format" and not isinstance(value, GeometryFormat):
            if isinstance(value, int):
                value = GeometryFormat(value)
            elif isinstance(value, str):
                value = GeometryFormat[value.upper().replace(" ", "_").replace("-", "_")]
            elif isinstance(value, _ext.CGeometryFormat):
                value = GeometryFormat(int(value))
            else:
                SansmicConfigError("geometry_format cannot be of type {}".format(type(value)))
        elif name == "units" and not isinstance(value, Units):
            if isinstance(value, int):
                value = Units(value)
            elif isinstance(value, str):
                value = Units[value.upper().replace(" ", "_").replace("-", "_")]
            else:
                SansmicConfigError("Units cannot be of type {}".format(type(value)))
        super().__setattr__(name, value)

    def _to_cmodel(self, prefix="temp"):
        """Create a C++ model object; in general, this should only be called internally."""
        cmodel = _ext.CModel(prefix)
        for stage_num, stage in enumerate(self.stages):
            if stage_num == 0:
                stage.set_initial_conditions = True
            cstage = stage._to_cstage()
            cstage.num_cells = self.num_cells
            cstage.geometry_format = _ext.CGeometryFormat(int(self.geometry_format))
            cstage.num_coallescing = self.coallescing_wells
            cstage.cavern_height = self.cavern_height
            cstage.ullage_standoff = self.ullage_standoff
            cstage.coallescing_well_separation = self.well_separation
            cstage.dissolution_factor = self.dissolution_factor
            cstage.insoluble_fraction = self.insolubles_ratio
            cstage.refDep = self.floor_depth
            cstage.depth = self.floor_depth
            if stage_num == 0:
                # Add the geometry data
                if self.geometry_format is GeometryFormat.RADIUS_LIST and isinstance(
                    self.geometry_data, str
                ):
                    radii = list()
                    with open(self.geometry_data, "r") as fin:
                        for line in fin.readlines():
                            if len(line.strip()) > 0:
                                radii.append(float(line.strip()))
                    cstage.radius_vector = radii
                elif isinstance(self.geometry_data, dict):
                    if "radii" in self.geometry_data:
                        tmp = self.geometry_data["radii"].copy()
                        tmp.insert(0, 0)
                        cstage.radius_vector = tmp
                        cstage.nData = len(self.geometry_data["radii"])
                    if "depths" in self.geometry_data:
                        tmp = self.geometry_data["depths"]
                        tmp.insert(0, 0)
                        cstage.depth_vector = tmp
                        cstage.nData = len(self.geometry_data["depths"])
                    if "volumes" in self.geometry_data:
                        tmp = self.geometry_data["volumes"]
                        tmp.insert(0, 0)
                        cstage.volume_vector = tmp
                        cstage.nData = len(self.geometry_data["volumes"])
            # add the stage to the c++ model
            cmodel.add_stage(cstage)
        return cmodel

    def new_stage(self, pos: int = None, **kwargs) -> StageDefinition:
        """Add a new stage to the scenario in the optionally specified `pos` position
        based on keyword arguments. Passes :attr:`Scenario.defaults` unless a separate
        `defaults` dictionary is passed as one of the keyword arguments.

        Parameters
        ----------
        pos : int or None, keyword only
            The position in the stages list to insert the stage,
            by default None which will append to the end.
        kwargs : keyword arguments
            Any valid keyword argument for a StageDefinition object can be passed.

        Returns
        -------
        StageDefinition
            new stage created from the keyword arguments which has already been
            inserted into the proper location
        """
        defaults = kwargs.pop("defaults", self.defaults)
        stage = StageDefinition(defaults=defaults, **kwargs)
        if pos is None:
            self.stages.append(stage)
        else:
            self.stages.insert(pos, stage)
        return stage

    def to_dict(self) -> Dict[str, Any]:
        """Convert the scenario definition to a dictionary"""
        ret = asdict(self)
        return ret

    def new_simulation(self, prefix="temp") -> "Simulator":
        """Create a new :class:`Simulator` object.

        Parameters
        ----------
        prefix : str
            The prefix to use when creating output files.
        """
        return Simulator(self, prefix)


class StageIterator:
    """Iterator over each stage within an open simulation.

    Examples
    --------
    >>> stages: StageIterator = sim.stages()
    >>> for stage in stages:
    ...     pass
    >>> res = sim.get_results()
    """

    def __init__(self, sim: "Simulator"):
        self._sim = sim
        self.__model_stages = self._sim.num_stages
        self.__current_stage = 0

    def __iter__(self):
        self.__model_stages = self._sim.num_stages
        self.__current_stage = 0
        return self

    def __next__(self):
        self._sim._has_run = True
        if self.__current_stage < self.__model_stages:
            self._sim._run_stage(self.__current_stage)
            self.__current_stage = self.__current_stage + 1
            return self.__current_stage
        else:
            raise StopIteration


class StepIterator:
    """Iterator over each step within an open simulation.

    Examples
    --------
    >>> steps: StepIterator = sim.steps()
    >>> for step in steps:
    ...     if step % 100 == 0:
    ...         step_res = sim.get_current_state()
    >>> res = sim.get_results()
    """

    def __init__(self, sim: "Simulator"):
        self._sim = sim
        self.__model_stages = self._sim.num_stages
        self.__current_stage = 0
        self.__current_step = 0

    def __iter__(self):
        self.__model_stages = self._sim.num_stages
        self.__current_stage = 0
        self.__current_step = 0
        return self

    def __next__(self):
        self._sim._has_run = True
        if self.__current_stage < self.__model_stages:
            cs = self._sim._run_step(self.__current_stage)
            self.__current_stage = cs
            self.__current_step = self.__current_step + 1
            return self.__current_step
        else:
            raise StopIteration

    @property
    def stage(self) -> int:
        """The current stage number."""
        return self.__current_stage


class Simulator:
    """The SANSMIC Python simulator.

    Parameters
    ----------
    scenario : Scenario
        The scenario to run
    prefix : str
        The output file prefix to use


    Examples
    --------
    Like Python file objects, the Simulator class can be used with a context
    manager -- the "with-as" syntax -- or by explicitly calling the
    :meth:`~Simulator.open` and :meth:`~Simulator.close` methods.

    """

    def __init__(self, scenario: Union[Scenario, _ext.CModel], prefix="temp"):
        self._scenario = None
        self._prefix = prefix
        self._cmodel = None
        if isinstance(scenario, Scenario):
            self._scenario = scenario
        else:
            raise TypeError("Invalid scenario type")
        self._has_run = False
        self._is_open = False
        self._is_initialized = False
        self._is_finalized = False

    def __enter__(self):
        self.open(self._prefix)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        pass

    def __del__(self):
        if self._is_open or self._cmodel is not None:
            try:
                self._cmodel.close_outfiles()
            except Exception as e:
                warnings.warn(e)
        del self._cmodel
        self._cmodel = None

    @property
    def is_open(self) -> bool:
        """Is this simulator attached to a C++ SANSMIC model?"""
        return self._is_open

    @property
    def current_stage(self) -> int:
        """The current stage of the simulation"""
        if not self._is_open or self._cmodel is None:
            raise RuntimeError("The simulation is not open")
        return self._cmodel._get_current_stage()

    @property
    def current_time(self) -> float:
        """The current time (within the stage) of the simulation"""
        if not self._is_open or self._cmodel is None:
            raise RuntimeError("The simulation is not open")
        return self._cmodel._get_current_time()

    @property
    def current_volume(self) -> float:
        """The current simulated cavern volume"""
        if not self._is_open or self._cmodel is None:
            raise RuntimeError("The simulation is not open")
        return self._cmodel._get_current_volume()

    @property
    def is_running(self):
        """Is the simulation mid-stage?"""
        if not self._is_open or self._cmodel is None:
            raise RuntimeError("The simulation is not open")
        return self._cmodel.is_running()

    @property
    def num_stages(self):
        """The total number of stages that have been defined"""
        if not self._is_open or self._cmodel is None:
            raise RuntimeError("The simulation is not open")
        return self._cmodel.num_stages()

    def _initialize(self):
        """Initialize the simulation (for iterators)"""
        if self._is_initialized:
            raise RuntimeError("The simulation has already been initialized")
        self._cmodel.initialize()
        self._cmodel.open_outfiles(False)
        self._is_initialized = True
        self._has_run = False
        self._is_finalized = False

    def _run_stage(self, stage_num: int):
        """Run the specified stage, all steps."""
        stage_num = self._cmodel.run_stage(stage_num)
        return stage_num

    def _run_step(self, stage_num: int):
        """Run the next timestep in the specified stage."""
        step_num = self._cmodel.run_step(stage_num)
        return step_num

    def _finalize(self):
        """Finalize the simulation (for iterators)"""
        if self._is_finalized:
            raise RuntimeError("The simulation has already been finalized")
        self._cmodel.close_outfiles()
        self._is_finalized = True
        self._is_initialized = True

    def open(self, prefix=None):
        """Open a simulation.

        Parameters
        ----------
        prefix : str
            The output filename prefix to use, by default None which
            uses the value specified when the Simulator was created.
        """
        if self._is_open or self._cmodel is not None:
            raise RuntimeError("The simulation is already open")
        if prefix is None:
            prefix = self._prefix
        else:
            self._prefix = prefix
        if self._cmodel is None and self._scenario is not None:
            self._cmodel = self._scenario._to_cmodel(prefix)
        self._is_open = True

    def close(self):
        """Close and garbage collect the C++ model object."""
        if not self._is_finalized and self._is_initialized:
            self._finalize()
        self._cmodel.close_outfiles()
        del self._cmodel
        self._cmodel = None
        self._is_open = False
        self._is_finalized = False
        self._is_initialized = False

    def run(self):
        """Run the complete simulation; requires the Simulator to have been opened first."""
        if self._cmodel is None:
            raise RuntimeError("The simulation is not open")
        self._is_finalized = False
        self._is_initialized = True
        self._cmodel.run()
        self._has_run = True
        self._is_initialized = True
        self._is_finalized = True

    def stages(self) -> StageIterator:
        """Get a generator that will run each stage in turn."""
        if not self._is_open or self._cmodel is None:
            raise RuntimeError("The simulation is not open")
        self._initialize()
        return StageIterator(self)

    def steps(self) -> StepIterator:
        """Get a generator that will run each step of each stage in turn."""
        if not self._is_open or self._cmodel is None:
            raise RuntimeError("The simulation is not open")
        self._initialize()
        return StepIterator(self)

    def get_current_state(self) -> "Results":
        """Get the current state of the model as a single-timestep results object."""
        if not self._is_open or self._cmodel is None:
            raise RuntimeError("The simulation is not open")
        data = self._cmodel._get_current_state()
        results = Results(data)
        return results

    def get_results(self) -> "Results":
        """Get the results for an entire simulation."""
        if self._cmodel is None:
            raise RuntimeError(
                "The simulation is already closed; results must be retrieved prior to closing the simulation"
            )
        if not self._has_run:
            raise RuntimeError("No results exist - the simulation has not been run/stages/steps.")
        data = self._cmodel._get_results()
        results = Results(data)
        return results


class Results:
    """
    Python container for the results of a sansmic simulation.

    Parameters
    ----------
    data : :class:`sansmic.library.libsansmic.CResults`
        results object from the C++ library

    """

    def __init__(self, data: _ext.CResults) -> None:
        self._data = data
        self._t_index = np.arange(1, len(self._data.t) + 1)
        self._n_index = np.arange(1, len(self._data.r[0]) + 1)

    def __repr__(self):
        return "<Results: from {:.3g} to {:.3g} d ({:d} stages)>".format(
            self._data.t[0] / 24.0,
            self._data.t[-1] / 24.0,
            max(self._data.stage) - min(self._data.stage) + 1,
        )

    def get_timeseries_data(self) -> pd.DataFrame:
        """Get the timeseries of summary results as a dataframe"""
        return pd.DataFrame.from_dict(
            {
                "time": self.t(),
                "days": self.t() / 24.0,
                "convergence": self.err(),
                "cavern volume": self.V_cavTot(),
                "insolubles volume": self.V_insolTot(),
                "insolubles vented": self.V_insolVent(),
                "water injected": self.V_injTot(),
                "oil injected": self.V_fillTot(),
                "output sg": self.sg_out(),
                "cavern sg": self.sg_cavAve(),
                "insolubles depth": self.z_insol(),
                "injection depth": self.z_inj(),
                "production depth": self.z_prod(),
                "plume depth": self.z_plm(),
                "obi depth": self.z_obi(),
                "production rate": self.Q_out(),
                "stage": self.stage(),
                "phase": self.phase(),
                "time step": self.dt(),
                "injection cell": self.injCell(),
                "production cell": self.prodCell(),
                "plume cell": self.plmCell(),
                "obi cell": self.obiCell(),
                "jet length": self.l_jet(),
                "jet velocity": self.u_jet(),
                "jet radius": self.r_jet(),
            }
        )

    def z_0(self) -> pd.Series:
        """node depth, in ft"""
        return pd.Series(self._data.z_0, index=self._n_index, name="depth")

    def h_0(self) -> pd.Series:
        """node height above original floor, in ft"""
        return pd.Series(self._data.h_0, index=self._n_index, name="height")

    def r_0(self) -> pd.Series:
        """initial cavern radius, in ft"""
        return pd.Series(self._data.r_0, index=self._n_index, name="radius")

    def t(self) -> pd.Series:
        """time of each row of time-based results"""
        return pd.Series(self._data.t, index=self._t_index, name="time")

    def stage(self) -> pd.Series:
        """stage number"""
        return pd.Series(self._data.stage, index=self._t_index, dtype=int, name="stage")

    def phase(self) -> pd.Series:
        """phase (0=static, 1=dynamic)"""
        return pd.Series(self._data.phase, index=self._t_index, dtype=int, name="phase")

    def injCell(self) -> pd.Series:
        """injection cell index"""
        return pd.Series(self._data.injCell, index=self._t_index, dtype=int, name="injection cell")

    def prodCell(self) -> pd.Series:
        """production cell index"""
        return pd.Series(
            self._data.prodCell, index=self._t_index, dtype=int, name="production cell"
        )

    def obiCell(self) -> pd.Series:
        """cell containing oil blanket interface"""
        return pd.Series(self._data.obiCell, index=self._t_index, dtype=int, name="obi cell")

    def plmCell(self) -> pd.Series:
        """cell containing plume stagnation level"""
        return pd.Series(self._data.plumCell, index=self._t_index, dtype=int)

    def err(self) -> pd.Series:
        """mass ballance error calculation"""
        return pd.Series(self._data.err, index=self._t_index, name="convergence")

    def z_obi(self) -> pd.Series:
        """depth of blanket interface"""
        return pd.Series(self._data.z_obi, index=self._t_index, name="obi depth")

    def z_inj(self) -> pd.Series:
        """depth of injection cell"""
        return pd.Series(self._data.z_inj, index=self._t_index, name="injection depth")

    def z_prod(self) -> pd.Series:
        """depth of production cell"""
        return pd.Series(self._data.z_prod, index=self._t_index, name="production depth")

    def z_plm(self) -> pd.Series:
        """depth of plume stagnation cell"""
        return pd.Series(self._data.z_plm, index=self._t_index, name="plume depth")

    def z_insol(self) -> pd.Series:
        """depth of top of insolubles"""
        return pd.Series(self._data.z_insol, index=self._t_index, name="insolubles depth")

    def h_insol(self) -> pd.Series:
        """height of insolubles above original floor"""
        return pd.Series(self._data.h_insol, index=self._t_index, name="insolubles height")

    def l_jet(self) -> pd.Series:
        """length of injection jet"""
        return pd.Series(self._data.l_jet, index=self._t_index, name="jet length")

    def r_jet(self) -> pd.Series:
        """velocity of injection jet"""
        return pd.Series(self._data.r_jet, index=self._t_index, name="jet radius")

    def u_jet(self) -> pd.Series:
        """radius of injection jet"""
        return pd.Series(self._data.u_jet, index=self._t_index, name="jet velocity")

    def V_injTot(self) -> pd.Series:
        """total injected water volume (bbl)"""
        return pd.Series(self._data.V_injTot, index=self._t_index, name="water injected")

    def V_fillTot(self) -> pd.Series:
        """total injected (+) or withdrawn (-) oil volume (bbl)"""
        return pd.Series(self._data.V_fillTot, index=self._t_index, name="oil injected")

    def V_cavTot(self) -> pd.Series:
        """total cavern volume (bbl)"""
        return pd.Series(self._data.V_cavTot, index=self._t_index, name="cavern volume")

    def V_insolTot(self) -> pd.Series:
        """total volume of insolubles (bbl)"""
        return pd.Series(self._data.V_insolTot, index=self._t_index, name="insolubles volume")

    def V_insolVent(self) -> pd.Series:
        """volume of insolubles vented (bbl)"""
        return pd.Series(self._data.V_insolVent, index=self._t_index, name="insolubles vented")

    def Q_out(self) -> pd.Series:
        """current brine exiting cavern (bbl/d)"""
        return pd.Series(self._data.Q_out, index=self._t_index, name="production rate")

    def sg_out(self) -> pd.Series:
        """current brine concentration exiting cavern (sg)"""
        return pd.Series(self._data.sg_out, index=self._t_index, name="output sg")

    def sg_cavAve(self) -> pd.Series:
        """current average concentration in brine (sg)"""
        return pd.Series(self._data.sg_cavAve, index=self._t_index, name="cavern sg")

    def dt(self) -> pd.Series:
        """the calculation timestep used"""
        return pd.Series(self._data.dt, index=self._t_index, name="time step")

    def r(self) -> pd.DataFrame:
        """current radius (ft)"""
        return pd.DataFrame(self._data.r, index=self._t_index, columns=self._n_index)

    def dr(self) -> pd.DataFrame:
        """change in radius since start of simulation (ft)"""
        return pd.DataFrame(self._data.dr_0, index=self._t_index, columns=self._n_index)

    def sg(self) -> pd.DataFrame:
        """current concentration (sg)"""
        return pd.DataFrame(self._data.sg, index=self._t_index, columns=self._n_index)

    def theta(self) -> pd.DataFrame:
        """current wall angle (deg)"""
        return pd.DataFrame(self._data.theta, index=self._t_index, columns=self._n_index)

    def Q_inj(self) -> pd.DataFrame:
        """current inflow (bbl/d)"""
        return pd.DataFrame(self._data.Q_inj, index=self._t_index, columns=self._n_index)

    def V(self) -> pd.DataFrame:
        """current volume of cell defined by node (bbl)"""
        return pd.DataFrame(self._data.V, index=self._t_index, columns=self._n_index)

    def f_dis(self) -> pd.DataFrame:
        """current dissolution factor, including adjustments"""
        return pd.DataFrame(self._data.f_dis, index=self._t_index, columns=self._n_index)

    def f_flag(self) -> pd.DataFrame:
        """current dissolution process adjustments flag"""
        return pd.DataFrame(
            self._data.f_flag, index=self._t_index, columns=self._n_index, dtype=int
        )

    def xincl(self) -> pd.DataFrame:
        """current wall angle adjustment factor"""
        return pd.DataFrame(self._data.xincl, index=self._t_index, columns=self._n_index)

    def amd(self) -> pd.DataFrame:
        """debug variable"""
        return pd.DataFrame(self._data.amd, index=self._t_index, columns=self._n_index)

    def D_coeff(self) -> pd.DataFrame:
        """current diffusion coefficient in use at node depth"""
        return pd.DataFrame(self._data.D_coeff, index=self._t_index, columns=self._n_index)

    def dC_dz(self) -> pd.DataFrame:
        """vertical concentration gradient (sg)"""
        return pd.DataFrame(self._data.dC_dz, index=self._t_index, columns=self._n_index)

    def C_old(self) -> pd.DataFrame:
        """previous step concentration (sg)"""
        return pd.DataFrame(self._data.C_old, index=self._t_index, columns=self._n_index)

    def C_new(self) -> pd.DataFrame:
        """current concentration (sg)"""
        return pd.DataFrame(self._data.C_new, index=self._t_index, columns=self._n_index)

    def dC_dt(self) -> pd.DataFrame:
        """change in concentration over time"""
        return pd.DataFrame(self._data.dC, index=self._t_index, columns=self._n_index)

    def dr_dt(self) -> pd.DataFrame:
        """current wall recession rate (ft)"""
        return pd.DataFrame(self._data.dr, index=self._t_index, columns=self._n_index)

    def C_plm(self) -> pd.DataFrame:
        """concentration within plume at node depth"""
        return pd.DataFrame(self._data.C_plm, index=self._t_index, columns=self._n_index)

    def u_plm(self) -> pd.DataFrame:
        """plume velocity at node depth"""
        return pd.DataFrame(self._data.u_plm, index=self._t_index, columns=self._n_index)

    def r_plm(self) -> pd.DataFrame:
        """plume radius at node depth"""
        return pd.DataFrame(self._data.r_plm, index=self._t_index, columns=self._n_index)


class _OutDataBlock:
    def __init__(self):
        self.t = None
        self.dt = None
        self.t_prime = None
        self.phase = None
        self.stage = None
        self.injCell = None
        self.prodCell = None
        self.obiCell = None
        self.plumCell = None
        self.z_inj = None
        self.z_prod = None
        self.z_obi = None
        self.z_plm = None
        self.l_jet = None
        self.r_jet = None
        self.u_jet = None
        self.idx = list()
        self.h = list()
        self.r = list()
        self.dr_0 = list()
        self.sg = list()
        self.theta = list()
        self.Q_inj = list()
        self.V = list()
        self.f_dis = list()
        self.f_flag = list()
        self.xincl = list()
        self.amd = list()
        self.D_coeff = list()
        self.dC_dz = list()
        self.C_old = list()
        self.C_new = list()
        self.dC = list()
        self.dr = list()
        self.C_plm = list()
        self.u_plm = list()
        self.r_plm = list()
        self.V_cavTot = None
        self.Q_out = None
        self.sg_out = None
        self.V_insolTot = None
        self.z_insol = None
        self.z_obi = None
        self.V_insolVent = None
        self.V_brine = None
        self.V_ullage = None
        self.V_usable = None
        self.err = None


class _OutputData(_OutDataBlock):
    def __init__(self):
        super().__init__()
        self.timeTotal = None
        self.r_0 = None
        self.z_0 = None
        self.h_0 = None
        self.V_injTot = None
        self.V_fillTot = None
        self.sg_cavAve = None


if not has_ext:
    _ext.CGeometryFormat = GeometryFormat
    _ext.CRunMode = SimulationMode
    _ext.CModel = Scenario
    _ext.CStage = StageDefinition



# coding: utf-8

import unittest
import sansmic

class TestModelRun(unittest.TestCase):
    def runTest(self):
        filename = 'old.dat'
        model = sansmic.read_scenario(filename)
        sansmic.write_scenario(model, 'new.toml')
        with model.new_simulation('new') as sim:
            sim.run()
            resPy = sim.get_results()
        resF = sansmic.io.read_classic_out_ddl('old')
        res_1 = resPy
        res_0 = resF
        resPy = resPy.get_timeseries_data()
        resF = resF.get_timeseries_data()
        resPy['modtime'] = (((resPy['time']*100)//10) / 10.0)
        resF['modtime'] = (((resF['time']*100)//10) / 10.0)
        resPy.set_index('modtime',inplace=True)
        resF.set_index('modtime',inplace=True)
        bothtimes = list(set(resF.index).intersection(set(resPy.index)))
        bothtimes.sort()
        # self.assertLess((resPy.loc[bothtimes,'cavern volume']-resF.loc[bothtimes,'cavern volume']).abs().max(),1e-4*resF['cavern volume'].max(), 'Volumes not within 0.01% through the entrie simulation')
        return resF, resPy

resF = resPy = None

if __name__ == '__main__':
    test = TestModelRun()
    resF, resPy = test.runTest()



# coding: utf-8
#
# Copyright (c) 2024 National Technology and Engineering Solutions of
# Sandia, LLC. Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# SPDX-License-Identifier: BSD-3-Clause.

"""Provide input and output functions for sansmic."""

import json
import logging
import os
import tomllib
import warnings
from dataclasses import fields
from enum import IntEnum

import numpy as np
import yaml

from .model import (
    GeometryFormat,
    Results,
    Scenario,
    SimulationMode,
    SansmicConfigError,
    StageDefinition,
    StopCondition,
    _OutDataBlock,
    _OutputData,
)

logger = logging.getLogger("sansmic")


def read_scenario(config_file: str, warn=True, strict=False) -> Scenario:
    """Load a sansmic scenario file.

    Parameters
    ----------
    config_file : str
        filename or full path to a new-style configuration file
    warn : bool, optional
        emit warnings if there are extra, unrecognized options in the file, by default True
    strict : bool, optional
        raise errors if there are extra, unrecognized options in the file, by default False

    Returns
    -------
    Scenario
        the sansmic scenario

    Raises
    ------
    RuntimeError
        if there is an unrecognized option and ``strict`` is True or if there are no
        stages defined.
    """
    logger.debug(
        "Attempting to read scenario {} in {} mode".format(
            config_file, "strict" if strict else "permissive"
        )
    )
    try:
        if config_file.lower().endswith(".json"):
            with open(config_file, "r") as fin:
                data = json.load(fin)
        elif config_file.lower().endswith(".yaml"):
            with open(config_file, "r") as fin:
                data = yaml.safe_load(fin)
        elif config_file.lower().endswith(".toml"):
            with open(config_file, "rb") as fin:
                data = tomllib.load(fin)
        elif config_file.lower().endswith(".dat"):
            with open(config_file, "r") as fin:
                data = read_dat(fin)
            return data
        else:
            with open(config_file, "rb") as fin:
                data = tomllib.load(fin)
    except:
        logger.error("Error reading scenario file {}".format(config_file))
        raise

    # Do error checking and remove extra fields so that the Scenario's __init__
    # does not have an exception
    scenario_fields = [f.name for f in fields(Scenario)]
    stage_fields = [f.name for f in fields(StageDefinition)]
    remove = list()
    for key in data.keys():
        if key not in scenario_fields:
            if strict:
                logger.error(
                    'Unrecognized option "{}" in scenario file options. Exiting'.format(key)
                )
                raise RuntimeError('Unrecognized option "{}" in scenario file options'.format(key))
            elif warn:
                logger.warn(
                    'Unrecognized option "{}" in scenario file options. Ignoring and continuing.'.format(
                        key
                    )
                )
            remove.append(key)
    for key in remove:
        data.pop(key)
    if "stages" not in data or len(data["stages"]) < 1:
        logger.error("No stages provided. Failed to load valid scenario.")
        raise RuntimeError("No stages provided.")
    for ct, stage in enumerate(data["stages"]):
        remove = list()
        for key in stage.keys():
            if key not in stage_fields:
                if strict:
                    logger.error(
                        'Unrecognized option "{}" in scenario file, stage {}. Exiting.'.format(
                            key, ct
                        )
                    )
                    raise RuntimeError(
                        'Unrecognized option "{}" in scenario file, stage {}'.format(key, ct)
                    )
                elif warn:
                    warnings.warn(
                        'Unrecognized option "{}" in scenario file, stage {}. Ignoring and continuing.'.format(
                            key, ct
                        )
                    )
                remove.append(key)
        for key in remove:
            stage.pop(key)
    return Scenario(**data)


def read_dat(str_or_buffer, *, ignore_errors=False) -> Scenario:
    """Read an old-style SANSMIC input DAT file.

    Parameters
    ----------
    str_or_buffer : str or stream
        The file or stream buffer to process.

    Returns
    -------
    Scenario
        the scenario derived from the DAT file

    """
    scenario = Scenario()
    with open(str_or_buffer, "r") if isinstance(str_or_buffer, str) else str_or_buffer as fin:
        prefix = fin.name
        scenario.title = "Converted from {}".format(prefix)
        first = True
        logger.debug("Reading old DAT format file: {}".format(prefix))
        while True:
            stage = StageDefinition()
            # Stage - block 1
            title = fin.readline().strip()
            if title == "END":
                break
            logger.debug(" Record 1: Stage title")
            logger.debug("  title         = {}".format(title))
            # Stage - block 2
            (
                num_cells,
                mode,
                print_interval,
                subsequent,
                iResetGeo,
                rest_duration,
                num_coallescing,
                geometry_format,
                stop_value,
            ) = (
                fin.readline().strip().split()
            )
            logger.debug(" Record 2: General")
            logger.debug("  ndiv          = {}".format(num_cells))
            logger.debug("  leachtype     = {}".format(mode))
            logger.debug("  iprint        = {}".format(print_interval))
            logger.debug("  repeat        = {}".format(subsequent))
            logger.debug("  resetgeo      = {}".format(iResetGeo))
            logger.debug("  iWait         = {}".format(rest_duration))
            logger.debug("  nco           = {}".format(num_coallescing))
            logger.debug("  idata         = {}".format(geometry_format))
            logger.debug("  ivol          = {}".format(stop_value))
            # Stage - block 3
            (
                cavern_height,
                injection_height,
                production_height,
                interface_height,
                ullage_standoff,
            ) = (
                fin.readline().strip().split()
            )
            logger.debug(" Record 3: Heights")
            logger.debug("  zmax          = {}".format(cavern_height))
            logger.debug("  zi            = {}".format(injection_height))
            logger.debug("  zp            = {}".format(production_height))
            logger.debug("  zb            = {}".format(interface_height))
            logger.debug("  zu            = {}".format(ullage_standoff))
            # Stage - block 4
            injection_rate = fin.readline().strip()
            logger.debug(" Record 4: Injection flow rates")
            logger.debug("  QI            = {}".format(injection_rate))
            # TODO: FIXME: handle injection tables? Or do we ignore for old
            # style inputs?
            #
            # Stage - block 5
            (
                inn_tbg_inside_radius,
                inn_tbg_outside_radius,
                out_csg_inside_radius,
                out_csg_outside_radius,
            ) = (
                fin.readline().strip().split()
            )
            logger.debug(" Record 5: Casing and tubing")
            logger.debug("  rpi           = {}".format(inn_tbg_inside_radius))
            logger.debug("  rpo           = {}".format(inn_tbg_outside_radius))
            logger.debug("  rcasi         = {}".format(out_csg_inside_radius))
            logger.debug("  rcaso         = {}".format(out_csg_outside_radius))
            # Stage - block 6
            injection_fluid_sg, cavern_sg = fin.readline().strip().split()
            logger.debug(" Record 6: Water and brine")
            logger.debug("  sginj         = {}".format(injection_fluid_sg))
            logger.debug("  sgcav         = {}".format(cavern_sg))
            # Stage - block 7
            timestep, injection_duration = fin.readline().strip().split()
            logger.debug(" Record 7: Timing")
            logger.debug("  dt            = {}".format(timestep))
            logger.debug("  tend          = {}".format(injection_duration))
            # Stage - block 8
            fill_rate, tDelay, coallescing_well_separation = fin.readline().strip().split()
            logger.debug(" Record 8: Oil fill rates")
            logger.debug("  qfil          = {}".format(fill_rate))
            logger.debug("  tdlay         = {}".format(tDelay))
            logger.debug("  sep           = {}".format(coallescing_well_separation))
            if float(tDelay) != 0:
                logger.warning("The TDLAY option should not be used. Ignoring.")
            # First stage - block 9
            if first:
                logger.debug("First stage initialization")
                logger.debug(" Record 9: Geometry")
                geometry_data = list()
                first = False
                geometry_format = GeometryFormat(int(geometry_format))
                if geometry_format == GeometryFormat.RADIUS_LIST:
                    radii = list()
                    for _ in range(int(num_cells) + 1):
                        r = float(fin.readline().strip())
                        radii.append(r)
                    geometry_data = dict(radii=radii)
                elif geometry_format == GeometryFormat.VOLUME_LIST:
                    ndata = int(fin.readline().strip())
                    depths = list()
                    volumes = list()
                    for _ in range(ndata):
                        z = float(fin.readline().strip())
                        depths.append(z)
                    for _ in range(ndata):
                        v = float(fin.readline().strip())
                        volumes.append(v)
                    geometry_data = dict(depths=depths, volumes=volumes)
                elif geometry_format == GeometryFormat.VOLUME_TABLE:
                    ndata = int(fin.readline().strip())
                    depths = list()
                    volumes = list()
                    for _ in range(ndata):
                        z, v = fin.readline().strip().split()
                        depths.append(float(z))
                        volumes.append(float(v))
                    geometry_data = dict(depths=depths, volumes=volumes)
                elif geometry_format == GeometryFormat.VOLUME_TABLE:
                    ndata = int(fin.readline().strip())
                    depths = list()
                    radii = list()
                    for _ in range(ndata):
                        z, v = fin.readline().strip().split()
                        depths.append(float(z))
                        radii.append(float(v))
                    geometry_data = dict(depths=depths, radii=radii)

                dissolution_factor, insoluble_fraction, refdep, depth = (
                    fin.readline().strip().split()
                )
                logger.debug(" Record 10: Miscellaneous")
                logger.debug("  zdis          = {}".format(dissolution_factor))
                logger.debug("  zfin          = {}".format(insoluble_fraction))
                logger.debug("  refdep        = {}".format(refdep))
                logger.debug("  depth         = {}".format(depth))
                if refdep != depth:
                    logger.warning("The REFDEP is no longer used, only DEPTH. Ignoring REFDEP.")
                if float(dissolution_factor) != 1.0:
                    logger.warning("The ZDIS should always be 1.0. This is a dangerous choice.")
                scenario.geometry_format = geometry_format
                scenario.geometry_data = geometry_data
                scenario.num_cells = int(num_cells)
                scenario.coallescing_wells = int(num_coallescing)
                scenario.well_separation = float(coallescing_well_separation)
                scenario.dissolution_factor = float(dissolution_factor)
                scenario.insolubles_ratio = float(insoluble_fraction)
                scenario.floor_depth = float(depth)
                scenario.cavern_height = float(cavern_height)
                scenario.ullage_standoff = float(ullage_standoff)
            else:
                logger.debug("Subsequent stage initialization")
            if not first and scenario.num_cells != int(num_cells) and not ignore_errors:
                logger.critical("Old-style DAT has changes in NDIV throughout the file")
                raise SansmicConfigError("Invalid data in DAT file: NDIV not constant.")
            elif not first and int(iResetGeo) != 0 and not ignore_errors:
                logger.critical(
                    "The RESETGEO flag (item 5 on line 2 of stage {}) is set to 1. This is not valid in this version of SANSMIC. Please correct this and rerun.".format(
                        len(scenario.stages) + 1
                    )
                )
                raise SansmicConfigError("Invalid data in DAT file: RESETGEO not 0")
            if not first and not bool(1 - int(subsequent)):
                logger.warning(
                    "The REPEAT option is supposed to turn off the cavern SG; initial cavern SG set to None."
                )
                cavern_sg = 0.0
            stage.title = title
            stage.simulation_mode = SimulationMode(int(mode))
            stage.save_frequency = int(print_interval)
            stage.set_initial_conditions = bool(1 - int(subsequent))
            stage.rest_duration = float(rest_duration)
            stage.stop_value = float(stop_value)
            if stage.stop_value < 0:
                stage.stop_condition = StopCondition.DEPTH
            if stage.stop_value > 0:
                stage.stop_condition = StopCondition.VOLUME
            stage.brine_injection_depth = float(injection_height)
            stage.brine_production_depth = float(production_height)
            stage.set_interface_level = float(interface_height)
            stage.brine_injection_rate = float(injection_rate)
            stage.inner_tbg_inside_diam = float(inn_tbg_inside_radius) * 2
            stage.inner_tbg_outside_diam = float(inn_tbg_outside_radius) * 2
            stage.outer_csg_inside_diam = float(out_csg_inside_radius) * 2
            stage.outer_csg_outside_diam = float(out_csg_outside_radius) * 2
            stage.brine_injection_sg = float(injection_fluid_sg)
            stage.set_cavern_sg = float(cavern_sg)
            stage.solver_timestep = float(timestep)
            stage.injection_duration = float(injection_duration)
            stage.product_injection_rate = float(fill_rate)
            scenario.stages.append(stage)
            logger.debug("Finished reading stage")
    return scenario


def write_scenario(scenario: Scenario, filename: str, *, redundant=False):
    """Write a new-style SANSMIC scenario file (preferred extension is .toml)"""
    sdict = scenario.to_dict()
    keys = [k for k in sdict.keys()]
    for k in keys:
        if not redundant and k == "dissolution_factor" and sdict[k] == 1.0:
            sdict[k] = None
        if not redundant and k == "coallescing_wells" and sdict[k] == 1:
            sdict[k] = None
            sdict["well_separation"] = None
        if sdict[k] is None:
            if redundant:
                sdict[k] = ""
            else:
                del sdict[k]
        elif isinstance(sdict[k], IntEnum):
            sdict[k] = sdict[k].name.lower().replace("_", "-")
    for ct, s in enumerate(sdict["stages"]):
        if not redundant and scenario.stages[ct].stop_condition == StopCondition.DURATION:
            del s["stop_condition"]
            del s["stop_value"]
        if scenario.stages[ct].set_initial_conditions is False:
            s["set_cavern_sg"] = None
            s["set_interface_level"] = None
            s["set_initial_conditions"] = None
        elif scenario.stages[ct].set_initial_conditions is None:
            s["set_cavern_sg"] = None
            if not s["set_interface_level"] and redundant:
                s["set_interface_level"] = 0
        if s.get("simulation_mode", None) == SimulationMode.WITHDRAWAL:
            s["product_injection_rate"] = None
        keys = [k for k in s.keys()]
        for k in keys:
            if not redundant and k in scenario.defaults and scenario.defaults.get(k, None) == s[k]:
                del s[k]
                continue
            if s[k] is None:
                if redundant:
                    s[k] = ""
                else:
                    del s[k]
            elif isinstance(s[k], IntEnum):
                s[k] = s[k].name.lower().replace("_", "-")
        if not redundant and ct == 0 and "set_initial_conditions" in s:
            del s["set_initial_conditions"]
    name, ext = os.path.splitext(filename)
    if not ext:
        filename = filename + ".toml"
        ext = '.toml'
    with open(filename, "w") as fout:
        if ext.lower() == ".toml":
            for k, v in sdict.items():
                if k in ["stages", "defaults"]:
                    continue
                elif k == "geometry_data" and isinstance(v, dict):
                    for k2, v2 in v.items():
                        fout.write("geometry_data.{} = {}\n".format(k2, v2))
                    continue
                if isinstance(v, bool):
                    v = str(v).lower()
                elif isinstance(v, str):
                    v = repr(v)
                elif isinstance(v, IntEnum):
                    v = repr(v.name.lower().replace("_", "-"))
                fout.write("{} = {}\n".format(k, v))

            if len(sdict["defaults"]) > 0:
                fout.write("\n[defaults]\n")
                for k2, v2 in sdict["defaults"].items():
                    fout.write("{} = {}\n".format(k2, v2))

            for s in sdict["stages"]:
                fout.write("\n[[stages]]\n")
                for k, v in s.items():
                    if isinstance(v, bool):
                        v = str(v).lower()
                    elif isinstance(v, str):
                        v = repr(v)
                    elif isinstance(v, IntEnum):
                        v = repr(v.name.lower().replace("_", "-"))
                    fout.write("{} = {}\n".format(k, v))
        elif ext.lower() == ".json":
            json.dump(sdict, fout)
        elif ext.lower() == ".yaml":
            yaml.dump(sdict, fout)
        else:
            logger.critical(
                'Unknown file extension {}. Valid extensions are "toml", "yaml" and "json" or none.'.format(
                    ext
                )
            )
            raise RuntimeError("Unknown file format for scenario output")


def read_classic_out_ddl(file_prefix):
    """
    Read in the data from ".out" and ".ddl" files from a classic SANSMIC run.

    Parameters
    ----------
    file_prefix : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """

    started = False
    dataline = False
    skip1 = False
    steps = list()
    step = None
    phaseline = False

    r_0 = list()
    z_0 = list()
    h_0 = list()

    with open(file_prefix + ".ddl", "r") as fddl:
        for ct, line in enumerate(fddl.readlines()):
            if ct == 0:
                continue
            words = line.split()
            idx = int(words[0])
            z = float(words[1])
            r = float(words[2])
            r_0.insert(0, r)
            z_0.insert(0, z)
            if idx == 1:
                break
    for z in z_0:
        h = z - z_0[0]
        h_0.append(h)

    with open(file_prefix + ".out", "r") as fout:
        for line in fout.readlines():
            words = line.split()
            if skip1:
                skip1 = False
                continue
            if not skip1 and len(words) == 0:
                continue
            if not started and words[0] == "TIME=":
                started = True
                step = _OutDataBlock()
            elif not started:
                continue
            if words[0] == "TIME=":
                step.t = float(words[1]) * 24.0
                step.dt = float(words[7]) * 24.0
                step.t_prime = float(words[11]) * 24.0
                phaseline = True
            elif phaseline and words[0] == "INJECTION":
                step.phase = 1
                step.stage = int(words[2])
                phaseline = False
            elif phaseline and words[0] == "WORKOVER":
                step.phase = 0
                step.stage = int(words[2])
                phaseline = False
            elif words[0] == "I(INJ),":
                step.injCell = int(words[4])
                step.prodCell = int(words[5])
                step.obiCell = int(words[6])
                step.plumCell = int(words[7])
            elif words[0] == "H(INJ),":
                step.z_inj = z_0[0] - float(words[4])
                step.z_prod = z_0[0] - float(words[5])
                # step.z_obi = z_0[0] - float(words[6])
                step.z_plm = z_0[0] - float(words[7])
            elif words[0] == "Ljet(ft),":
                step.l_jet = float(words[4])
                step.r_jet = float(words[5])
                step.u_jet = float(words[6])
                skip1 = True
            elif words[0] == "HEIGHT":
                skip1 = True
                dataline = True
            elif dataline:
                cellid = int(line[0:5])
                if cellid == 1:
                    dataline = False
                    skip1 = True
                step.idx.insert(0, cellid)
                step.h.insert(0, float(line[5:16]))
                step.r.insert(0, float(line[16:27]))
                step.dr_0.insert(0, float(line[27:38]))
                step.sg.insert(0, float(line[38:49]))
                step.theta.insert(0, float(line[49:60]))
                step.Q_inj.insert(0, float(line[60:71]))
                step.V.insert(0, float(line[71:82]))
                step.f_dis.insert(0, float(line[82:93]))
                step.f_flag.insert(0, int(line[93:98]))
                step.xincl.insert(0, float(line[98:109]))
                step.amd.insert(0, float(line[109:120]))
                step.D_coeff.insert(0, float(line[120:131]))
                step.dC_dz.insert(0, float(line[131:142]))
                step.C_old.insert(0, float(line[142:153]))
                step.C_new.insert(0, float(line[153:164]))
                step.dC.insert(0, float(line[164:175]))
                step.dr.insert(0, float(line[175:186]))
                step.C_plm.insert(0, float(line[186:197]))
                step.u_plm.insert(0, float(line[197:208]))
                step.r_plm.insert(0, float(line[208:]))
            elif words[0] == "TOTAL":
                step.V_cavTot = float(line[25:36])
            elif words[0] == "BRINE" and words[1] == "OUT":
                step.Q_out = float(line[25:36])
            elif words[0] == "OUTLET":
                step.sg_out = float(line[25:36])
            elif words[0] == "VOLUME":
                step.V_insolTot = float(line[25:36])
            elif words[0] == "INSOL":
                step.z_insol = z_0[0] - float(line[25:36])
            elif words[0] == "BLANKET":
                step.z_obi = z_0[0] - float(line[25:36])
            elif words[0] == "VOL":
                step.V_insolVent = float(line[25:36])
            elif words[0] == "BRINE" and words[1] == "VOLUME":
                step.V_brine = float(line[25:36])
            elif words[0] == "ULLAGE":
                step.V_ullage = float(line[25:36])
            elif words[0] == "USEABLE":
                step.V_usable = float(line[25:36])
            elif words[0] == "CFAC":
                step.err = float(line[25:36])
            elif line.startswith("----------"):
                # TODO: add datablock to list
                steps.append(step)
                started = False
            else:
                print("unrecognized line: ", line)
    combo = _OutputData()
    combo.z_0 = z_0
    combo.h_0 = h_0
    combo.r_0 = r_0
    combo.V_injTot = list()
    combo.V_fillTot = list()
    combo.sg_cavAve = list()

    for step in steps:
        combo.V_injTot.append(np.nan)
        combo.V_fillTot.append(np.nan)
        combo.sg_cavAve.append(np.nan)
        for key in step.__dict__.keys():
            if not hasattr(combo, key):
                continue
            else:
                tmp = getattr(combo, key)
                if tmp is None:
                    setattr(combo, key, list())
                if key == "t":
                    val = step.t + step.t_prime
                else:
                    val = getattr(step, key)
                getattr(combo, key).append(val)

    res = Results(combo)
    return res
